---@type Tinkr, Bastion
local Tinkr, Bastion = ...

-- Create an EventManager class
---@class Bastion.EventManager
---@field frame Frame
---@field events table<string, { [number]: fun(...) }>
---@field eventHandlers table<string, { [number]: fun(...) }>
---@field wowEventHandlers table<WowEvent, { [number]: fun(...) }>
---@field selfCombatEventHandlers table<string, { [number]: fun(...) }>
---@field CombatEventHandlers table<string, { [number]: fun(...) }>
---@field playerGUID string|boolean
local EventManager = {
    events = {},
    eventHandlers = {},
    wowEventHandlers = {},
    selfCombatEventHandlers = {},
    CombatEventHandlers = {},
}
EventManager.__index = EventManager

-- Constructor
---@return Bastion.EventManager
function EventManager:New()
    local self = setmetatable({}, EventManager)
    self.events = {}
    self.eventHandlers = {}
    self.wowEventHandlers = {}
    self.selfCombatEventHandlers = {}
    self.CombatEventHandlers = {}
    self.playerGUID = UnitGUID("player") or false

    -- Frame for wow events
    self.frame = CreateFrame("Frame")

    self.frame:SetScript("OnEvent", function(f, event, ...)
        if self.wowEventHandlers[event] then
            for _, callback in ipairs(self.wowEventHandlers[event]) do
                callback(...)
            end
        end
    end)

    self:RegisterWoWEvent("COMBAT_LOG_EVENT_UNFILTERED", function()
        self:CLEUHandler(CombatLogGetCurrentEventInfo())
    end)

    return self
end

-- Register an event
---@param event string
---@param handler fun(...)
---@return nil
function EventManager:RegisterEvent(event, handler)
    if not self.events[event] then
        self.events[event] = {}
    end

    table.insert(self.events[event], handler)
end

-- Register a wow event
---@param events WowEvent | WowEvent[]
---@param handler fun(...)
function EventManager:RegisterWoWEvent(events, handler)
    if type(events) == "string" then
        events = { events }
    end

    for _, event in ipairs(events) do
        if not self.wowEventHandlers[event] then
            self.wowEventHandlers[event] = {}
            self.frame:RegisterEvent(event)
        end

        table.insert(self.wowEventHandlers[event], handler)
    end
end

-- Trigger an event
---@param event string
---@param ... any
---@return nil
function EventManager:TriggerEvent(event, ...)
    if self.events[event] then
        for _, handler in pairs(self.events[event]) do
            handler(...)
        end
    end
end

---@param subevents string | string[]
---@param handler fun(...)
function EventManager:RegisterSelfCombatEvent(subevents, handler)
    if type(subevents) == "string" then
        subevents = { subevents }
    end
    for _, subevent in ipairs(subevents) do
        if not self.selfCombatEventHandlers[subevent] then
            self.selfCombatEventHandlers[subevent] = {}
        end

        table.insert(self.selfCombatEventHandlers[subevent], handler)
    end
end

---@param subevents string | string[]
---@param handler fun(...)
function EventManager:RegisterCombatEvent(subevents, handler)
    if type(subevents) == "string" then
        subevents = { subevents }
    end
    for _, subevent in ipairs(subevents) do
        if not self.CombatEventHandlers[subevent] then
            self.CombatEventHandlers[subevent] = {}
        end

        table.insert(self.CombatEventHandlers[subevent], handler)
    end
end

---@param timestamp number
---@param subevent string
---@param ... any
function EventManager:CLEUHandler(timestamp, subevent, ...)
    if self.selfCombatEventHandlers[subevent] and self.playerGUID and select(2, ...) == self.playerGUID then
        for _, callback in ipairs(self.selfCombatEventHandlers[subevent]) do
            callback(timestamp, subevent, ...)
        end
    end
    if self.CombatEventHandlers[subevent] then
        for _, callback in ipairs(self.CombatEventHandlers[subevent]) do
            callback(timestamp, subevent, ...)
        end
    end
end

return EventManager
